#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <stdint.h>

#include <math.h>
#include <float.h>

#include <string.h>

#include "clFFT.h"

#define BLOCK 256


typedef struct {
    int width;
    int height;
    clFFT_Complex *elements;
} Matrix;

typedef struct {
    int width;
    int height;
    float *reals;
    float *imags;
} MatrixSplit;


//used for copying Matrices in GPU storage
typedef struct {
    cl_int width;
    cl_int height;
    cl_mem elements;
} GPUMatrix;

#define MRI_CHANNELS 1

static void
get_event_times (cl_event event, unsigned long *start_wait, unsigned long *execution)
{
    cl_ulong queued, start, end;

    OCL_CHECK_ERROR (clGetEventProfilingInfo (event, CL_PROFILING_COMMAND_QUEUED, sizeof (cl_ulong), &queued, NULL));
    OCL_CHECK_ERROR (clGetEventProfilingInfo (event, CL_PROFILING_COMMAND_START, sizeof (cl_ulong), &start, NULL));
    OCL_CHECK_ERROR (clGetEventProfilingInfo (event, CL_PROFILING_COMMAND_END, sizeof (cl_ulong), &end, NULL));

    *start_wait = start - queued;
    *execution = end - start;
}

static void
convertSplitToInterleaved(size_t noElements, float *reals, float *imags, clFFT_Complex *interleaved)
{
    for (unsigned i = 0; i < noElements; ++i)
    {
        interleaved[i].real = reals[i];
        interleaved[i].imag = imags[i];
    }
}

static void
convertInterleavedToSplitMatrix(Matrix const *interleaved, MatrixSplit *split)
{
    split->width = interleaved->width;
    split->height = interleaved->height;
    int size = split->width * split->height;
    split->reals = (float *) malloc (sizeof(float) * size);
    split->imags = (float *) malloc (sizeof(float) * size);
    for (int i = 0; i < size; ++i)
    {
        split->reals[i] = interleaved->elements[i].real;
        split->imags[i] = interleaved->elements[i].imag;
    }
}

// reads complex numbers from binary file stored in row major order
static int
getDataFromSplitBinary (FILE *f, Matrix *ret)
{

    int wh[2];   
    fread (wh, sizeof(int), 2, f);

    ret->width = wh[0];
    ret->height = wh[1];
    printf ("w:%d  h:%d\n", ret->width, ret->height);
    size_t size = ret->width * ret->height;
    ret->elements = (clFFT_Complex *) malloc(sizeof(clFFT_Complex)*size);
    float *lmns = (float *) malloc (sizeof (float) * 2 * size);
    
    fread (lmns, sizeof(float), size * 2, f);

    convertSplitToInterleaved (size, lmns, (lmns + size), ret->elements);

    free (lmns);

    return 1;
}

static int
readFilterFromFile (int isSmooth, int n, float **filter, int *filterlength)
{
    char filtname[64];
    if (isSmooth != 0)
        sprintf(filtname, "filt/bin/LowPass%sDecim%d_filt.bin", "Smooth", n);
    else
        sprintf(filtname, "filt/bin/LowPass%sDecim%d_filt.bin", "Sharp", n);

    FILE *f = fopen(filtname, "r");

    if (f == NULL)
    {
        printf ("Error opening file!\n%s\n", filtname);
        exit (1);
    }

    fseek (f, 0L, SEEK_END);
    size_t size = ftell (f) / sizeof(float);
    rewind (f);

    *filter = (float*) malloc (size * sizeof (float));

    fread (*filter, sizeof(float), size, f);
    *filterlength = size;

    return 1;
}

static void
writeSplitMatrixToBinaryFile(FILE *f, MatrixSplit * const matrix)
{
    fwrite (matrix->reals, sizeof (float), matrix->width * matrix->height, f);
    fwrite (matrix->imags, sizeof (float), matrix->width * matrix->height, f);
}


static void
tranformFromKSpace (Matrix *A, Matrix *result,
             int noSlices,
             const cl_context context,
             const cl_kernel fftshiftKernel,
             const cl_kernel absKernel,
             const cl_command_queue queue)
{
    cl_int errcode;
    // measure event queue
    cl_event event;
    // times in micro seconds
    unsigned long wait;
    unsigned long execution;

    unsigned long waitFFTShift[2];
    unsigned long executionFFTShift[2];
    unsigned long computeAbsSpectrum;
    unsigned long readFromDevice;

    // setup measurement of host times
    GTimer *timer;
    timer = g_timer_new ();
    double wallClock = 0.0;

    // FFT dimension
    clFFT_Dim3 n = { A->width, A->height, 1};
    int length = n.x * n.y;

    /*Matrix C;
    C.width = A->width;
    C.height = A->height;
    C.elements = (clFFT_Complex *) malloc(sizeof(clFFT_Complex)*length);*/

    result->width = A->width;
    result->height = A->height;
    result->elements = (clFFT_Complex *) malloc(sizeof(clFFT_Complex)*length);

    cl_mem tempBuffer = clCreateBuffer (context, CL_MEM_READ_WRITE,
                                   length * sizeof(float) * 2, NULL, &errcode);
    OCL_CHECK_ERROR (errcode);

    GPUMatrix d_C;
    d_C.width = result->width;
    d_C.height = result->height;
    d_C.elements = clCreateBuffer (context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
                                   length * sizeof(float) * 2 * noSlices, result->elements, &errcode);

    OCL_CHECK_ERROR (errcode);

    GPUMatrix d_A;
    d_A.width = A->width;
    d_A.height = A->height;

    d_A.elements = clCreateBuffer (context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
                                   length * sizeof(float) * 2 * noSlices, A->elements, &errcode);

    OCL_CHECK_ERROR (errcode);

    clFFT_Plan fftPlan = clFFT_CreatePlan (context, n,
                                           clFFT_2D, clFFT_InterleavedComplexFormat,
                                           &errcode);

    OCL_CHECK_ERROR (errcode);

    // inverse transform
    OCL_CHECK_ERROR (clFFT_ExecuteInterleaved (queue, fftPlan, noSlices, clFFT_Inverse, d_A.elements,
                              d_C.elements, 0, NULL, NULL));



    // fftshift
    size_t globalWorkSize[] = { d_C.width, d_C.height };
    // swap along x axis
    cl_int dim = 0;
    cl_uint i = 0;
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.width), 
                    (void*) &d_C.width));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.height), 
                    (void*) &d_C.height));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.elements), 
                    (void*) &d_C.elements));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (tempBuffer), 
                    (void*) &tempBuffer));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (dim), 
                    (void*) &dim));

    g_timer_start (timer);
    OCL_CHECK_ERROR (clEnqueueNDRangeKernel (queue, fftshiftKernel, 2, NULL, 
                            globalWorkSize, NULL, 
                            0, NULL, &event));

    clWaitForEvents (1, &event);
    g_timer_stop (timer);
    get_event_times (event, &wait, &execution);
    waitFFTShift[0] = wait;
    executionFFTShift[0] = execution;
    wallClock = g_timer_elapsed (timer, NULL);

    printf("fftshift x axis took %e on host\n", wallClock);

   // swap along y axis
    dim = 1;
    i = 0;
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.width), 
                    (void*) &d_C.width));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.height), 
                    (void*) &d_C.height));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (tempBuffer), 
                    (void*) &tempBuffer));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (d_C.elements), 
                    (void*) &d_C.elements));
    OCL_CHECK_ERROR (clSetKernelArg(fftshiftKernel, i++, sizeof (dim), 
                    (void*) &dim));

    g_timer_start (timer);
    OCL_CHECK_ERROR (clEnqueueNDRangeKernel (queue, fftshiftKernel, 2, NULL, 
                            globalWorkSize, NULL, 
                            0, NULL, &event));

    clWaitForEvents (1, &event);
    g_timer_stop (timer);

    get_event_times (event, &wait, &execution);
    waitFFTShift[1] = wait;
    executionFFTShift[1] = execution;

    wallClock = g_timer_elapsed (timer, NULL);

    printf("fftshift y axis took %e on host\n", wallClock);

    // power spectral density
    i = 0;
    OCL_CHECK_ERROR (clSetKernelArg(absKernel, i++, sizeof (d_C.width), 
                    (void*) &d_C.width));
    OCL_CHECK_ERROR (clSetKernelArg(absKernel, i++, sizeof (d_C.height), 
                    (void*) &d_C.height));
    OCL_CHECK_ERROR (clSetKernelArg(absKernel, i++, sizeof (d_C.elements), 
                    (void*) &d_C.elements));

    OCL_CHECK_ERROR (clEnqueueNDRangeKernel (queue, absKernel, 2, NULL, 
                            globalWorkSize, NULL, 
                            0, NULL, &event));

    clWaitForEvents (1, &event);
    g_timer_stop (timer);
    get_event_times (event, &wait, &execution);
    computeAbsSpectrum = execution;

    g_timer_start (timer);
    // Read C from device memory
    errcode = clEnqueueReadBuffer (queue, d_C.elements, CL_TRUE, 0,
                                   length * sizeof(float) * 2, result->elements,
                                   0, NULL, &event);
    OCL_CHECK_ERROR (errcode);
    clWaitForEvents (1, &event);
    get_event_times (event, &wait, &execution);
    readFromDevice = wait + execution;

    wallClock = g_timer_elapsed (timer, NULL);


    printf("%lu\n%lu\n%lu\n%lu\n%lu\n%lu\n", 
            waitFFTShift[0], waitFFTShift[1],
            executionFFTShift[0], executionFFTShift[1],
            computeAbsSpectrum,
            readFromDevice);

    g_timer_destroy (timer);


    clFFT_DestroyPlan (fftPlan);
    OCL_CHECK_ERROR (clReleaseEvent (event));
    OCL_CHECK_ERROR (clReleaseMemObject (tempBuffer));
    OCL_CHECK_ERROR (clReleaseMemObject (d_A.elements));
    OCL_CHECK_ERROR (clReleaseMemObject (d_C.elements));
}

static void
filterData (int decimFactor, int isSmooth, Matrix * const input, Matrix *decimOutput,
             const cl_context context,
             const cl_kernel convKernel,
             const cl_command_queue queue)
{

    cl_int errcode;
    // measure event queue
    cl_event event;
    // times in micro seconds
    unsigned long wait;
    unsigned long execution;


    float *filter;
    int filterlength = 0;
    readFilterFromFile (isSmooth, decimFactor, &filter, &filterlength);

    int length = input->width * input->height;

    OCL_CHECK_ERROR (errcode);
    cl_int d_decimFactor = decimFactor;
    cl_int d_filterlength = filterlength;
    cl_mem d_filter = clCreateBuffer (context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
                                   filterlength * sizeof(float), filter, &errcode);
    OCL_CHECK_ERROR (errcode);

    GPUMatrix d_C;
    d_C.width = input->width;
    d_C.height = input->height;
    d_C.elements = clCreateBuffer (context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
                                   length * sizeof(float) * 2, input->elements, &errcode);

    OCL_CHECK_ERROR (errcode);

    decimOutput->width = d_C.width;
    decimOutput->height = d_C.height;
    decimOutput->elements = (clFFT_Complex *) malloc(sizeof(clFFT_Complex)*length);

    GPUMatrix d_filtered;
    d_filtered.width = d_C.width;
    d_filtered.height = d_C.height;
    d_filtered.elements = clCreateBuffer (context, CL_MEM_READ_WRITE,
                                   length * sizeof(float) * 2, NULL, &errcode);
    OCL_CHECK_ERROR (errcode);

/*    size_t globalWorkSize[] = { d_filtered.width, d_filtered.height };
    cl_uint i = 0;
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filtered.width), 
                    (void*) &d_filtered.width));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_C.height), 
                    (void*) &d_C.height));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_C.elements), 
                    (void*) &d_C.elements));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filtered.elements), 
                    (void*) &d_filtered.elements));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_decimFactor), 
                    (void*) &d_decimFactor));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filter), 
                    (void*) &d_filter));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filterlength), 
                    (void*) &d_filterlength));

   
    OCL_CHECK_ERROR (clEnqueueNDRangeKernel (queue, convKernel, 2, NULL, 
                            globalWorkSize, NULL, 
                            0, NULL, &event));*/

    size_t globalWorkSize[] = { d_C.width, d_C.height };
    size_t localWorkSize[] = { 256, 1 };
    cl_uint i = 0;
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_C.width), 
                    (void*) &d_C.width));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_C.height), 
                    (void*) &d_C.height));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_decimFactor), 
                    (void*) &d_decimFactor));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filterlength), 
                    (void*) &d_filterlength));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_C.elements), 
                    (void*) &d_C.elements));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filtered.elements), 
                    (void*) &d_filtered.elements));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, sizeof (d_filter), 
                    (void*) &d_filter));

    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, localWorkSize[0] * localWorkSize[1] * sizeof (float), 
                    NULL));
    OCL_CHECK_ERROR (clSetKernelArg(convKernel, i++, localWorkSize[0] * localWorkSize[1] * sizeof (float), 
                    NULL));
   
    OCL_CHECK_ERROR (clEnqueueNDRangeKernel (queue, convKernel, 2, NULL, 
                            globalWorkSize, localWorkSize, 
                            0, NULL, &event));

    clWaitForEvents (1, &event);
    get_event_times (event, &wait, &execution);
    printf("Convolution: \n waited for device %lu us:\n time to execute: %lu nanoseconds\n", wait, execution);


    errcode = clEnqueueReadBuffer (queue, d_filtered.elements, CL_TRUE, 0,
                                   length * sizeof(float) * 2, decimOutput->elements,
                                   0, NULL, &event);
    clWaitForEvents (1, &event);
    get_event_times (event, &wait, &execution);
    printf("Read decimated k space from device: \n waited for device %lu us:\n time to execute: %lu us\n", wait, execution);


    OCL_CHECK_ERROR (errcode);

    OCL_CHECK_ERROR (clReleaseEvent (event));
    OCL_CHECK_ERROR (clReleaseMemObject (d_C.elements));
    OCL_CHECK_ERROR (clReleaseMemObject (d_filtered.elements));

}


int
main (int argc, const char **argv)
{
    // declare variable for OpenCL environment
    OclPlatform *ocl;
    cl_context context;
    cl_command_queue *queues;
    cl_int errcode;

    // auxiliary program with custom kernels;
    cl_program program;
    cl_kernel absKernel;
    cl_kernel fftshiftKernel;
    cl_kernel convolutionKernelInterleaved;

    ocl = ocl_new_from_args (argc, argv, CL_QUEUE_PROFILING_ENABLE);
    context = ocl_get_context (ocl);
    queues = ocl_get_cmd_queues (ocl);

    program = ocl_create_program_from_file (ocl, "ffthelp.cl", NULL, &errcode);
    OCL_CHECK_ERROR (errcode);
    absKernel = clCreateKernel (program, "absComplexInterleaved", &errcode);
    OCL_CHECK_ERROR (errcode);
    fftshiftKernel = clCreateKernel (program, "fftshift", &errcode);
    OCL_CHECK_ERROR (errcode);
    convolutionKernelInterleaved = clCreateKernel (program, "convolutionLocal", &errcode);
    OCL_CHECK_ERROR (errcode);

    //create Matrices
    Matrix k_Space, decimk_Space, absimg;

    FILE *bin = fopen ("ksd.bin", "r");
    if (bin == NULL)
    {
        printf ("Error opening file!\n");
        exit (1);
    }

    getDataFromSplitBinary (bin, &k_Space);
    fclose (bin);

    filterData (8, 1, &k_Space, &decimk_Space, 
                        context, convolutionKernelInterleaved, queues[0]);

    tranformFromKSpace (&decimk_Space, &absimg, MRI_CHANNELS, context,
                        fftshiftKernel, absKernel, queues[0]);

    MatrixSplit absimgSplit;
    convertInterleavedToSplitMatrix (&decimk_Space, &absimgSplit);

    FILE *f = fopen("imgocl.bin", "wb");
    if (f == NULL)
    {
        printf("Error opening file!\n");
        exit(1);
    }

    writeSplitMatrixToBinaryFile (f, &absimgSplit);
    fclose (f);

    ocl_free (ocl);

    return 0;
}
